/***********************************************************************************************************************
 *
 * blueBill Mobile - open source birdwatching
 * ==========================================
 *
 * Copyright (C) 2009, 2010 by Tidalwave s.a.s. (http://www.tidalwave.it)
 * http://bluebill.tidalwave.it/mobile/
 *
 ***********************************************************************************************************************
 *
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
 * an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
 * specific language governing permissions and limitations under the License.
 *
 ***********************************************************************************************************************
 *
 * $Id: TaxonSoundFactSheetController.java,v 8656dd4f49ae 2010/07/16 17:21:40 fabrizio $
 *
 **********************************************************************************************************************/
package it.tidalwave.bluebill.mobile.taxonomy.factsheet.sound;

import it.tidalwave.util.As;
import it.tidalwave.bluebill.mobile.taxonomy.factsheet.FactSheetProvider;
import it.tidalwave.bluebill.mobile.taxonomy.factsheet.SoundObservationNode;
import it.tidalwave.semantic.io.GraphDeserializer;
import it.tidalwave.semantic.io.GraphUnmarshaller;
import it.tidalwave.semantic.io.impl.DefaultGraphDeserializer;
import it.tidalwave.semantic.io.impl.GraphUnmarshallerImpl;
import it.tidalwave.util.Id;
import org.openrdf.model.Graph;
import javax.annotation.CheckForNull;
import javax.annotation.Nonnegative;
import javax.annotation.Nonnull;
import javax.inject.Provider;
import java.beans.PropertyChangeEvent;
import java.beans.PropertyChangeListener;
import java.util.Arrays;
import java.util.List;
import java.io.InputStream;
import java.io.IOException;
import org.openide.util.NbBundle;
import it.tidalwave.util.AsException;
import it.tidalwave.util.Identifiable;
import it.tidalwave.util.Removable;
import it.tidalwave.util.logging.Logger;
import it.tidalwave.netbeans.util.Locator;
import it.tidalwave.mobile.util.Downloadable;
import it.tidalwave.mobile.util.Downloadable.Status;
import it.tidalwave.mobile.io.MasterFileSystem;
import it.tidalwave.mobile.media.MediaPlayer;
import it.tidalwave.mobile.media.MediaPlayer.Controller;
import it.tidalwave.mobile.media.Media;
import it.tidalwave.bluebill.observation.Observation;
import it.tidalwave.bluebill.taxonomy.Taxon;
import it.tidalwave.bluebill.observation.ObservationSet;
import it.tidalwave.bluebill.mobile.network.NetworkingPreferences;
import it.tidalwave.bluebill.mobile.taxonomy.factsheet.TaxonFactSheetControllerSupport;
import static it.tidalwave.mobile.util.Downloadable.Status.*;

/***********************************************************************************************************************
 *
 * @author  Fabrizio Giudici
 * @version $Id: $
 *
 **********************************************************************************************************************/
public abstract class TaxonSoundFactSheetController extends TaxonFactSheetControllerSupport<TaxonSoundFactSheetUI>
  {
    private static final String CLASS = TaxonSoundFactSheetController.class.getName();
    private static final Logger logger = Logger.getLogger(CLASS);

    private final static double MEGA = 1024.0 * 1024.0;
    /* package */ final static long MIN_FREE_SPACE = 20L * 1024 * 1024;
    private final static List<String> WATCHED_PROPERTIES = Arrays.asList(Downloadable.PROP_STATUS, Downloadable.PROP_DOWNLOAD_PROGRESS);

    private enum Operation
      {
        DOWNLOAD
          {
            public void run (final @Nonnull Downloadable downloadable)
              {
                downloadable.download();
              }
          },
        REFRESH
          {
            public void run (final @Nonnull Downloadable downloadable)
              {
                downloadable.refresh();
              }
          };

        public abstract void run (@Nonnull Downloadable downloadable);
      }

    public static enum DefaultSoundAction implements DefaultAction<TaxonSoundFactSheetController>
      {
        DO_NOTHING
          {
            public void run (final @Nonnull TaxonSoundFactSheetController controller, final @Nonnull Media media)
              {
              }
          },
        DOWNLOAD
          {
            public void run (final @Nonnull TaxonSoundFactSheetController controller, final @Nonnull Media media)
              {
                controller.downloadMedia(media);
              }
          },
        PLAY
          {
            public void run (final @Nonnull TaxonSoundFactSheetController controller, final @Nonnull Media media)
              {
                controller.playMedia(media);
              }
          },
        STOP
          {
            public void run (final @Nonnull TaxonSoundFactSheetController controller, final @Nonnull Media media)
              {
                controller.stopPlayer();
              }
          };
      }

    @CheckForNull
    /* package */ Controller controller;

    @Nonnull
    private final Provider<MediaPlayer> mediaPlayer = Locator.createProviderFor(MediaPlayer.class);

    @Nonnull
    private final Provider<MasterFileSystem> masterFileSystem = Locator.createProviderFor(MasterFileSystem.class);

    @CheckForNull
    private Media playingMedia;

    /*******************************************************************************************************************
     *
     *
     ******************************************************************************************************************/
    /* package */ final PropertyChangeListener mediaChangeListener = new PropertyChangeListener()
      {
        public void propertyChange (final @Nonnull PropertyChangeEvent event)
          {
            notifyStatusChanged();
          }
      };

    /*******************************************************************************************************************
     *
     *
     ******************************************************************************************************************/
    /* package */ final PropertyChangeListener downloadablePropertyChangeListener = new PropertyChangeListener()
      {
        public void propertyChange (final @Nonnull PropertyChangeEvent event)
          {
//            logger.info("propertyChange(%s)", event);

            if (WATCHED_PROPERTIES.contains(event.getPropertyName()) && (event.getOldValue() != event.getNewValue()))
              {
                notifyStatusChanged();

                if (event.getPropertyName().equals(Downloadable.PROP_STATUS))
                  {
                    switch ((Status)event.getNewValue())
                      {
                        case DOWNLOADED:
                          final double freeSpace = masterFileSystem.get().getExternalFileSystem().getFreeSpace() / MEGA;
                          ui.notify(NbBundle.getMessage(TaxonSoundFactSheetController.class, "downloadComplete", freeSpace));
                          ((Downloadable)event.getSource()).removePropertyChangeListener(downloadablePropertyChangeListener);
                          break;

                        case BROKEN:
                          ui.notify(NbBundle.getMessage(TaxonSoundFactSheetController.class, "downloadFailed"));
                          ((Downloadable)event.getSource()).removePropertyChangeListener(downloadablePropertyChangeListener);
                          break;

                        case DOWNLOADING:
                        case NOT_DOWNLOADED:
                        case OBSOLETE:
                        case QUEUED:
                      }
                  }
              }
          }
      };

    /*******************************************************************************************************************
     *
     *
     ******************************************************************************************************************/
    public TaxonSoundFactSheetController (final @Nonnull TaxonSoundFactSheetUI ui, 
                                          final @Nonnull Taxon taxon,
                                          final @Nonnull FactSheetProvider factSheetProvider)
      {
        super(ui, taxon, factSheetProvider);
      }

    /*******************************************************************************************************************
     *
     *
     ******************************************************************************************************************/
    public void disposePlayer()
      {
        if (controller != null)
          {
            controller.removePropertyChangeListener(mediaChangeListener);
            controller.dispose();
            controller = null;
          }
      }

    /*******************************************************************************************************************
     *
     *
     ******************************************************************************************************************/
    public boolean isDownloadable (final @Nonnull Media media)
      {
        return isStatus(media, NOT_DOWNLOADED, OBSOLETE, BROKEN);
      }

    /*******************************************************************************************************************
     *
     *
     ******************************************************************************************************************/
    public boolean isRefreshable (final @Nonnull Media media)
      {
        return isStatus(media, DOWNLOADED, OBSOLETE, BROKEN);
      }

    /*******************************************************************************************************************
     *
     *
     ******************************************************************************************************************/
    public boolean isDeletable (final @Nonnull Media media)
      {
        return isStatus(media, DOWNLOADED, OBSOLETE);
      }

    /*******************************************************************************************************************
     *
     *
     ******************************************************************************************************************/
    public boolean isPlayable (final @Nonnull Media media)
      {
        return isStatus(media, DOWNLOADED, OBSOLETE);
      }

    /*******************************************************************************************************************
     *
     *
     ******************************************************************************************************************/
    public boolean isPlaying (final @Nonnull Media media)
      {
        return isPlayable(media) && (controller != null) && controller.isPlaying() && (media == playingMedia);
      }

    /*******************************************************************************************************************
     *
     *
     ******************************************************************************************************************/
    public boolean hasWebPage (final @Nonnull Observation observation)
      {
        try
          {
            observation.as(Identifiable.class).getId();
            return true;
          }
        catch (AsException e)
          {
            return false;
          }
      }

    /*******************************************************************************************************************
     *
     *
     ******************************************************************************************************************/
    public void downloadMedia (final @Nonnull Media media)
      {
        if (isDownloadable(media))
          {
            downloadMediaWithConfirmation(media, Operation.DOWNLOAD);
          }
      }

    /*******************************************************************************************************************
     *
     *
     ******************************************************************************************************************/
    public void refreshMedia (final @Nonnull Media media)
      {
        if (isRefreshable(media))
          {
            downloadMediaWithConfirmation(media, Operation.REFRESH);
          }
      }

    /*******************************************************************************************************************
     *
     *
     ******************************************************************************************************************/
    public void deleteMedia (final @Nonnull Media media)
      {
        if (isDeletable(media))
          {
            final String title   = NbBundle.getMessage(TaxonSoundFactSheetController.class, "confirmDeletionTitle");
            final String message = NbBundle.getMessage(TaxonSoundFactSheetController.class, "confirmDeletionMessage");

            ui.getCommonUITasks().alertDialog(title, message, new Runnable()
              {
                public void run()
                  {
                    try
                      {
                        media.as(Removable.class).remove();
                        final double freeSpace = masterFileSystem.get().getExternalFileSystem().getFreeSpace() / MEGA;
                        ui.notify(NbBundle.getMessage(TaxonSoundFactSheetController.class, "deletionCompleted", freeSpace));
                      }
                    catch (Exception e)
                      {
                        ui.notify(NbBundle.getMessage(TaxonSoundFactSheetController.class, "deletionFailed"));
                        logger.warning("Can't delete media %s", e);
                        logger.throwing("deleteMedia()", CLASS, e);
                      }

                    // Needed because at this point the PropertyChangeListener has been detached.
                    // Maybe it would be better to keep the listener attached, implement a dispose() method that
                    // detaches it at the end.
                    notifyStatusChanged();
                  }
              });
          }
      }

    /*******************************************************************************************************************
     *
     *
     ******************************************************************************************************************/
    public void playMedia (final @Nonnull Media media)
      {
        if (isPlayable(media))
          {
            try
              {
                disposePlayer();
                playingMedia = media;
                controller = mediaPlayer.get().play(media);
                controller.addPropertyChangeListener(mediaChangeListener);
                ui.notifyPlayMedia(controller);
                notifyStatusChanged();
              }
            catch (Exception e)
              {
                ui.notify(NbBundle.getMessage(TaxonSoundFactSheetController.class, "playFailed"));
                logger.warning("Can't play sound %s", e);
                logger.throwing("downloadMedia()", CLASS, e);
              }
          }
      }

    /*******************************************************************************************************************
     *
     *
     ******************************************************************************************************************/
    public void stopPlayer()
      {
        disposePlayer();
        notifyStatusChanged();
      }

    /*******************************************************************************************************************
     *
     *
     ******************************************************************************************************************/
    @Nonnull
    public DefaultSoundAction getDefaultAction (final @Nonnull Media media)
      {
        if (isPlaying(media))
          {
            return DefaultSoundAction.STOP;
          }
        else if (isPlayable(media))
          {
            return DefaultSoundAction.PLAY;
          }
        else if (isDownloadable(media))
          {
            return DefaultSoundAction.DOWNLOAD;
          }
        else
          {
            return DefaultSoundAction.DO_NOTHING;
          }
      }

    /*******************************************************************************************************************
     *
     *
     ******************************************************************************************************************/
    @Nonnull
    public Media getMedia (final @Nonnegative int index)
      {
        return getObservation(index).as(Media.class);
      }

    /*******************************************************************************************************************
     *
     *
     ******************************************************************************************************************/
    @Nonnull
    public Observation getObservation (final @Nonnegative int index)
      {
        return ((SoundObservationNode)nodes.get(index)).getObservation();
      }

    /*******************************************************************************************************************
     *
     *
     ******************************************************************************************************************/
    protected abstract void notifyStatusChanged();

    /*******************************************************************************************************************
     *
     *
     ******************************************************************************************************************/
    @Nonnull
    protected ObservationSet loadObservationSet (final @Nonnull InputStream is, final @Nonnull String mimeType)
      throws IOException
      {
        try
          {
            final GraphDeserializer deserializer = new DefaultGraphDeserializer();
            final Graph graph = deserializer.read(is, mimeType);
            final GraphUnmarshaller unmarshaller = new GraphUnmarshallerImpl();
            final Id type = new Id("http://www.tidalwave.it/rdf/observation/2010/07/01#ObservationSet");
            return (ObservationSet)unmarshaller.unmarshal(graph, type);
          }
        catch (Exception e)
          {
            throw new RuntimeException(e);
//            throw new IOException(e.toString()); // Java 5 / Android compatibility
          }
      }
    
    /*******************************************************************************************************************
     *
     *
     ******************************************************************************************************************/
    private void downloadMediaWithConfirmation (final @Nonnull Media media, final @Nonnull Operation operation)
      {
        final String title   = NbBundle.getMessage(TaxonSoundFactSheetController.class, "confirmDownloadTitle");
        final String message = NbBundle.getMessage(TaxonSoundFactSheetController.class, "confirmDownloadMessage");

        if (Locator.find(NetworkingPreferences.class).isNetworkConnectionAllowed())
          {
            doDownloadMedia(media, operation);
          }
        else
          {
            ui.getCommonUITasks().alertDialog(title, message, new Runnable()
              {
                public void run()
                  {
                    doDownloadMedia(media, operation);
                  }
              });
          }
      }

    /*******************************************************************************************************************
     *
     *
     ******************************************************************************************************************/
    /* package */ void doDownloadMedia (final @Nonnull Media media, final @Nonnull Operation operation)
      {
        if (!masterFileSystem.get().getExternalFileSystem().isWritable())
          {
            final String title   = NbBundle.getMessage(TaxonSoundFactSheetController.class, "downloadFailed");
            final String message = NbBundle.getMessage(TaxonSoundFactSheetController.class, "diskNotMountedMessage");
            ui.getCommonUITasks().alertDialog(title, message, null);
          }
        else if(masterFileSystem.get().getExternalFileSystem().getFreeSpace() <= MIN_FREE_SPACE)
          {
            final String title   = NbBundle.getMessage(TaxonSoundFactSheetController.class, "downloadFailed", MIN_FREE_SPACE / MEGA);
            final String message = NbBundle.getMessage(TaxonSoundFactSheetController.class, "diskFullMessage", MIN_FREE_SPACE / MEGA);
            ui.getCommonUITasks().alertDialog(title, message, null);
          }
        else
          {
            final Downloadable downloadable = media.as(Downloadable.class);
            downloadable.addPropertyChangeListener(downloadablePropertyChangeListener);
            operation.run(downloadable);
            ping();
          }
      }

    /*******************************************************************************************************************
     *
     *
     ******************************************************************************************************************/
    private boolean isStatus (final @Nonnull Media media, final @Nonnull Status ... status)
      {
        final Downloadable downloadable = media.as(Downloadable.class);
        return Arrays.asList(status).contains(downloadable.getStatus());
      }

    protected void ping()
      {
      }
  }
